from inspect import signature
from unittest.mock import patch

from dash.exceptions import PreventUpdate
from django.test import TestCase

from tom_alerts_dash.brokers.scimma import SCIMMADashBroker
from tom_alerts_dash.tests.factories import create_scimma_alert, SiderealTargetFactory


class TestSCIMMADashBroker(TestCase):

    def setUp(self):
        self.broker = SCIMMADashBroker()

        self.test_target = SiderealTargetFactory.create()
        self.test_alerts = [create_scimma_alert() for i in range(0, 5)]

    def test_flatten_dash_alerts(self):
        test_alert = create_scimma_alert()
        flattened_alerts = self.broker.flatten_dash_alerts([test_alert])
        for key in ['alert_identifier', 'counterpart_identifier', 'ra', 'dec', 'rank', 'comments']:
            self.assertIn(key, flattened_alerts[0])
        for key in ['topic', 'test_bad_key']:
            self.assertNotIn(key, flattened_alerts[0])  # Test that no unwanted attributes are included

        expected_url = f'https://gracedb.ligo.org/superevents/{test_alert["message"]["event_trig_num"]}/view/'
        self.assertDictContainsSubset(
            {'alert_identifier': f'[{test_alert["alert_identifier"]}]({expected_url})',
             'counterpart_identifier': test_alert['extracted_fields']['counterpart_identifier'],
             'ra': test_alert['right_ascension_sexagesimal'],
             'dec': test_alert['declination_sexagesimal'],
             'rank': test_alert['message']['rank'],
             'comments': test_alert['extracted_fields']['comment_warnings'],
             'alert': test_alert},
            flattened_alerts[0])

    def test_callback_partial_cone_search(self):
        with self.assertRaises(PreventUpdate):
            self.broker.callback(1, 20, '', '', '100', None, None, None, None)

    @patch('tom_scimma.scimma.SCIMMABroker._request_alerts')
    def test_callback_full_cone_search(self, mock_request_alerts):
        mock_request_alerts.return_value = {'results': self.test_alerts}
        alerts = self.broker.callback(1, 20, '', '', '100', '100', '100', None, None)

        self.assertDictContainsSubset({'cone_search': '100,100,100'}, mock_request_alerts.call_args.args[0])
        for key in ['alert_identifier', 'counterpart_identifier', 'ra', 'dec', 'rank', 'comments']:
            self.assertIn(key, alerts[0])
        for key in ['topic', 'test_bad_key']:
            self.assertNotIn(key, alerts[0])  # Test that no unwanted attributes are included

    def test_validate_filters(self):
        errors = self.broker.validate_filters(1, 20, '', '', '100', None, None, None, None, [])
        self.assertIn('All of RA, Dec, and Radius are required for a cone search.', errors[0].children)

    def test_callback_parameters_match_inputs(self):
        """Test that callback function has the same number of parameters as the inputs."""
        callback_num_params = len(signature(self.broker.callback).parameters)
        inputs = self.broker.get_callback_inputs()
        self.assertEqual(callback_num_params, len(inputs))

    def test_inputs_match_filters(self):
        callback_inputs = self.broker.get_callback_inputs()
        dash_filters = self.broker.get_dash_filters()
        for callback_input in callback_inputs:
            if callback_input.component_property in ['page_current', 'page_size']:
                continue
            with self.subTest():
                found = False
                for row in dash_filters.children:
                    for column in row.children:
                        for input_obj in column:
                            if input_obj == callback_input.component_id:
                                found = True
                self.assertTrue(found)



from http import HTTPStatus
from unittest.mock import patch

from dash.dependencies import Input
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import dash_html_components as dhc
from django.test import override_settings, TestCase
from django.urls import reverse

from tom_alerts_dash.alerts import GenericDashBroker
from tom_alerts_dash.dash_apps.query_list_app import broker_selection_callback, create_broker_container, create_targets
from tom_targets.models import Target


class TestDashBroker(GenericDashBroker):
    name = 'Test Broker'

    def callback(self, page_current, page_size, test_input):
        return [
            {'test_key': test_input}
        ]

    def fetch_alerts(self):
        return []

    def get_callback_inputs(self):
        inputs = super().get_callback_inputs()
        return inputs + [Input('test-input', 'value')]

    def get_dash_filters(self):
        return dhc.Div([
            dcc.Input(
                id='test-input',
                type='text'
            )
        ])

    def get_dash_columns(self):
        return [{'id': 'test_key', 'name': 'Test Key', 'type': 'text'}]

    def to_generic_alert(self):
        return


class TestDashViews(TestCase):

    def setUp(self):
        pass

    def test_broker_query_list_view(self):
        """Test that the tom_alerts_dash ListView is used rather than the tom_alerts ListView."""
        response = self.client.get(reverse('tom_alerts_dash:list'))
        self.assertContains(response, 'Browse Alerts')

    def test_broker_query_browse_view(self):
        """Test that the BrokerQueryBrowseView loads."""
        response = self.client.get(reverse('tom_alerts_dash:browse'))
        self.assertEqual(response.status_code, HTTPStatus.OK)


@override_settings(TOM_ALERT_DASH_CLASSES=['tom_alerts_dash.tests.tests.TestDashBroker'])
class TestQueryListApp(TestCase):

    def setUp(self):
        pass

    def test_create_broker_container(self):
        broker_container = create_broker_container('Test Broker')
        for key in ['create-targets-btn-Test Broker', 'alerts-table-Test Broker']:
            self.assertIn(key, broker_container)
        self.assertEqual(broker_container.style, {'display': 'none'})

    @patch('tom_alerts_dash.dash_apps.query_list_app.reverse')
    @patch('tom_alerts_dash.tests.tests.TestDashBroker.to_target')
    def test_create_targets(self, mock_to_target, mock_reverse):
        mock_reverse.return_value = 'http://localhost:8000/targets/1/'

        params = [('', [1, 2], [], 'Test Broker', []), (1, [], [], 'Test Broker', [])]
        for param in params:
            with self.subTest():
                with self.assertRaises(PreventUpdate):
                    create_targets(*param)

        mock_to_target.side_effect = lambda target: target
        rows = [{'alert': Target(id=1, name='test1')}, {'alert': None}]
        with self.subTest():
            messages = create_targets(1, [0], rows, 'Test Broker', [])
            self.assertIn('Successfully created ', messages[0].children)

        with self.subTest():
            messages = create_targets(1, [1], rows, 'Test Broker', [])
            self.assertIn('Unable to create target from alert.', messages[0].children)

    def test_broker_selection_callback(self):
        params = [('', ''), ('Test Broker', 'Test Broker')]
        for param in params:
            with self.subTest():
                with self.assertRaises(PreventUpdate):
                    broker_selection_callback(*param)

        with self.subTest():
            callback_return_values = broker_selection_callback('Test Broker', '')
            self.assertEqual('Test Broker', callback_return_values[0])
            self.assertEqual('Test Broker Alerts', callback_return_values[1].children)
            self.assertDictEqual({'display': 'block'}, callback_return_values[2])

        with self.subTest():
            callback_return_values = broker_selection_callback('Other Broker', 'Test Broker')
            self.assertDictEqual({'display': 'none'}, callback_return_values[2])



from datetime import datetime
import logging

from dash.dependencies import Input
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import dash_html_components as dhc
import dash_core_components as dcc

from tom_alerts_dash.alerts import GenericDashBroker
from tom_scimma.scimma import SCIMMABroker, SCIMMAQueryForm

logger = logging.getLogger(__name__)

GRACE_DB_URL = 'https://gracedb.ligo.org'


class SCIMMADashBroker(SCIMMABroker, GenericDashBroker):

    def callback(self, page_current, page_size, event_trigger_number, keyword, cone_ra, cone_dec, cone_radius,
                 start_date, end_date):
        """
        SCIMMA-specific callback function for BrokerQueryBrowseView. Queries SCIMMA based on parameters from DataTable
        inputs. The callback will not modify any bound components if one or more, but not all, cone search inputs are
        submitted.

        :param page_current: Currently selected page
        :type page_current: int

        :param page_size: Page size for pagination. Not currently used.
        :type page_size: int

        :param event_trigger_number: Event trigger number to search for
        :type event_trigger_number: string

        :param keyword: Keyword to search for
        :type keyword: string

        :param cone_ra: Right Ascension to use in cone search
        :type cone_ra: string

        :param cone_dec: Declination to use in cone search
        :type cone_dec: string

        :param cone_radius: Radius to use in cone search, in degrees
        :type cone_radius: string

        :param start_date: Earliest date to filter by
        :param start_date: string

        :param end_date: Latest date to filter by
        :param end_date: string

        :returns: list of flattened alerts
        :rtype: list of dicts

        :raises: PreventUpdate exception if some but not all cone search parameters are submitted
        """
        logger.info('Entering SCIMMA callback...')
        errors = self.validate_filters(page_current, page_size, event_trigger_number, keyword, cone_ra, cone_dec,
                                       cone_radius, start_date, end_date, [])
        if errors:
            raise PreventUpdate

        cone_search = ''
        if any([cone_ra, cone_dec, cone_radius]):
            if all([cone_ra, cone_dec, cone_radius]):
                cone_search = ','.join([cone_ra, cone_dec, cone_radius])
            else:
                raise PreventUpdate

        form = SCIMMAQueryForm({
            'query_name': 'SCIMMA Dash Query',
            'broker': self.name,
            'keyword': keyword,
            'cone_search': cone_search,
            'event_trigger_number': event_trigger_number,
            'alert_timestamp_after': start_date,
            'alert_timestamp_before': end_date,
        })
        form.is_valid()

        parameters = form.cleaned_data
        parameters['topic'] = 1  # form isn't valid with both topic and event trigger number, so this circumvents that
        parameters['page'] = page_current + 1  # Dash pagination is 0-indexed, but Skip is 1-indexed
        parameters['page_size'] = page_size if page_size else 20  # 20 is the Dash default page size
        alerts = self._request_alerts(parameters)['results']
        return self.flatten_dash_alerts(alerts)

    def get_callback_inputs(self):
        """
        Returns SCIMMA-specific inputs used to trigger callback function.

        :returns: list of inputs corresponding to dash filters
        :rtype: list
        """
        inputs = super().get_callback_inputs()
        inputs += [
            Input('keyword', 'value'),
            Input('event-trigger-number', 'value'),
            Input('scimma-ra', 'value'),
            Input('scimma-dec', 'value'),
            Input('scimma-radius', 'value'),
            Input('date-filter', 'start_date'),
            Input('date-filter', 'end_date'),
        ]
        return inputs

    def get_dash_filters(self):
        """
        Returns SCIMMA-specific filter inputs layout

        :returns: layout of Dash input components
        :rtype: dash_html_components.Div
        """
        filters = dhc.Div([
            dbc.Row([
                dbc.Col(dcc.Input(
                    id='keyword',
                    type='text',
                    placeholder='Keyword Search',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='event-trigger-number',
                    type='text',
                    placeholder='LVC Trigger Number',
                    debounce=True
                ), width=3)
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dcc.Input(
                    id='scimma-ra',
                    type='text',
                    placeholder='Right Ascension',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='scimma-dec',
                    type='text',
                    placeholder='Declination',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='scimma-radius',
                    type='text',
                    placeholder='Radius',
                    debounce=True
                ), width=3)
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dcc.DatePickerRange(
                    id='date-filter',
                    min_date_allowed=datetime(2020, 1, 1),
                    initial_visible_month=datetime.now(),
                    clearable=True
                ))
            ], style={'padding-bottom': '10px'}, justify='start'),
        ])
        return filters

    def get_dash_columns(self):
        """
        Returns SCIMMA-specific Dash DataTable columns.

        :returns: columns for display
        :rtype: list of dicts
        """
        return [
            {'id': 'alert_identifier', 'name': 'Alert Identifier', 'type': 'text', 'presentation': 'markdown'},
            {'id': 'counterpart_identifier', 'name': 'Counterpart Identifier', 'type': 'text'},
            {'id': 'ra', 'name': 'Right Ascension', 'type': 'text'},
            {'id': 'dec', 'name': 'Declination', 'type': 'text'},
            {'id': 'rank', 'name': 'Rank', 'type': 'text'},
            {'id': 'comments', 'name': 'Comments', 'type': 'text'}
        ]

    def flatten_dash_alerts(self, alerts):
        """
        Transforms alerts returned by SCIMMA into a Dash DataTable format. Adds an embedded link to the original alert.
        Converts decimal degrees to sexagesimal.

        :param alerts: list of alerts from SCIMMA
        :type alerts: list of dicts

        :returns: flattened alerts
        :rtype: list of dicts
        """
        flattened_alerts = []
        for alert in alerts:
            url = f'{GRACE_DB_URL}/superevents/{alert["message"]["event_trig_num"]}/view/'
            flattened_alerts.append({
                'alert_identifier': f'[{alert["alert_identifier"]}]({url})',
                'counterpart_identifier': alert['extracted_fields']['counterpart_identifier'],
                'ra': alert['right_ascension_sexagesimal'],
                'dec': alert['declination_sexagesimal'],
                'rank': alert['message']['rank'],
                'comments': alert['extracted_fields']['comment_warnings'],
                'alert': alert
            })
        return flattened_alerts

    def validate_filters(self, page_current, page_size, event_trigger_number, keyword, cone_ra, cone_dec, cone_radius,
                         start_date, end_date, errors_state):
        """
        Validates the input filters for SCIMMA. Returns an error if one, but not all, of RA, Dec, and radius are
        submitted for cone search. Returns any errors generated by form validation.

        :param page_current: The page number for the paginated alerts to display
        :type page_current: int

        :param page_size: The page size used for the pagination
        :type page_size: int

        :param event_trigger_number: Event trigger number to search for
        :type event_trigger_number: string

        :param keyword: Keyword to search for
        :type keyword: string

        :param cone_ra: Right Ascension to use in cone search
        :type cone_ra: string

        :param cone_dec: Declination to use in cone search
        :type cone_dec: string

        :param cone_radius: Radius to use in cone search, in degrees
        :type cone_radius: string

        :param start_date: Earliest date to filter by
        :param start_date: string

        :param end_date: Latest date to filter by
        :param end_date: string

        :param errors_state: The currently displayed errors relating to filters
        :type errors_state: list of dbc.Alert objects

        :returns: errors from validation of filters
        :rtype: list of dbc.Alert objects
        """
        errors = []

        cone_search = ''
        if any([cone_ra, cone_dec, cone_radius]):
            if all([cone_ra, cone_dec, cone_radius]):
                cone_search = ','.join([cone_ra, cone_dec, cone_radius])
            else:
                errors.append('All of RA, Dec, and Radius are required for a cone search.')

        form = SCIMMAQueryForm({
            'query_name': 'SCIMMA Dash Query',
            'broker': self.name,
            'keyword': keyword,
            'cone_search': cone_search,
            'event_trigger_number': event_trigger_number,
            'alert_timestamp_after': start_date,
            'alert_timestamp_before': end_date,
        })
        form.is_valid()

        for field, field_errors in form.errors.items():
            for field_error in field_errors.get_json_data():
                errors.append(f'{field}: {field_error["message"]}')

        for error in errors:
            errors_state.append(dbc.Alert(error, dismissable=True, is_open=True, duration=5000, color='warning'))

        return errors_state



from datetime import datetime
from inspect import signature
from unittest.mock import patch

from dash.exceptions import PreventUpdate
from django.test import TestCase

from tom_alerts_dash.brokers.alerce import ALeRCEDashBroker
from tom_alerts_dash.tests.factories import create_alerce_alert, SiderealTargetFactory


class TestALeRCEDashBroker(TestCase):

    def setUp(self):
        self.broker = ALeRCEDashBroker()
        self.test_target = SiderealTargetFactory.create()

        self.mock_classifiers = [
            {'classifier_name': 'lc_classifier', 'classifier_version': 'hierarchical_random_forest_1.0.0',
             'classes': ['SNIa', 'SNIbc']},
            {'classifier_name': 'lc_classifier_top', 'classifier_version': 'hierarchical_random_forest_1.0.0',
             'classes': ['Transient', 'Stochastic', 'Periodic']},
            {'classifier_name': 'stamp_classifier', 'classifier_version': 'stamp_classifier_1.0.4',
             'classes': ['SN', 'AGN', 'VS', 'asteroid', 'bogus']}
        ]

    @patch('tom_alerts.brokers.alerce.ALeRCEQueryForm._get_classifiers')
    def test_flatten_dash_alerts(self, mock_get_classifiers):
        mock_get_classifiers.return_value = self.mock_classifiers

        test_alert_late_class = create_alerce_alert(meanra=60, meandec=120, classifier='lc_classifier',
                                                    class_name='SNIa', probability=0.98765)
        test_alert_early_class = create_alerce_alert(firstmjd=59196.441261574075, classifier='stamp_classifier',
                                                     class_name='SN', probability=0.54321)
        alerts = [test_alert_late_class,
                  test_alert_early_class]
        flattened_alerts = self.broker.flatten_dash_alerts(alerts)
        for key in ['oid', 'meanra', 'meandec', 'discovery_date', 'classifier', 'class', 'probability']:
            self.assertIn(key, flattened_alerts[0])
        for key in ['last_mjd', 'test_bad_key']:
            self.assertNotIn('last_mjd', flattened_alerts[0])  # Test that no unwanted attributes are included
        self.assertDictContainsSubset(
            {'oid': f'[{test_alert_late_class["oid"]}](https://alerce.online/object/{test_alert_late_class["oid"]})',
             'meanra': '04:00:0.000',
             'meandec': '+120:00:0.000',
             'class': 'SNIa',
             'classifier': 'lc_classifier',
             'probability': '0.9877'},
            flattened_alerts[0]
        )

        self.assertDictContainsSubset(
            {'discovery_date': datetime(2020, 12, 13, 10, 35, 25),
             'class': 'SN',
             'classifier': 'stamp_classifier',
             'probability': '0.5432'},
            flattened_alerts[-1])

    def test_callback_no_button_click(self):
        with self.assertRaises(PreventUpdate):
            self.broker.callback(1, 20, None, None, None, None, None, None, None, None, None)

        with self.assertRaises(PreventUpdate):
            self.broker.dash_button_clicks = 1
            self.broker.callback(1, 20, None, None, None, None, None, None, None, None, 1)

    @patch('tom_alerts.brokers.alerce.ALeRCEQueryForm._get_classifiers')
    @patch('tom_alerts.brokers.alerce.ALeRCEBroker._request_alerts')
    def test_callback(self, mock_request_alerts, mock_get_classifiers):
        mock_get_classifiers.return_value = self.mock_classifiers
        test_alerts = [create_alerce_alert() for i in range(0, 5)]
        mock_request_alerts.return_value = {'items': test_alerts}

        alerts = self.broker.callback(1, 20, None, None, None, None, None, None, None, None, 10)
        self.assertEqual(len(alerts), len(test_alerts))
        self.assertEqual(self.broker.dash_button_clicks, 10)

    def test_callback_parameters_match_inputs(self):
        """Test that callback function has the same number of parameters as the inputs."""
        callback_num_params = len(signature(self.broker.callback).parameters)
        inputs = self.broker.get_callback_inputs()
        self.assertEqual(callback_num_params, len(inputs))

    def test_validate_filters(self):
        with self.assertRaises(PreventUpdate):
            self.broker.validate_filters(1, 20, None, None, None, None, None, 100, '', '', None, [])

        errors = self.broker.validate_filters(1, 20, None, None, None, None, None, 100, '', '', 10, [])
        self.assertIn('__all__: All of RA, Dec, and Search Radius must be included to execute a cone search.',
                      errors[0].children)

    def test_inputs_match_filters(self):
        callback_inputs = self.broker.get_callback_inputs()
        dash_filters = self.broker.get_dash_filters()
        for callback_input in callback_inputs:
            if callback_input.component_property in ['page_current', 'page_size']:
                continue
            with self.subTest():
                found = False
                for row in dash_filters.children:
                    for column in row.children:
                        for input_obj in column:
                            if input_obj == callback_input.component_id:
                                found = True
                self.assertTrue(found)



"""
Django settings for tom_alerts_dash project.

Generated by 'django-admin startproject' using Django 3.1.3.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

import os
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'b@*iv%&mhkyewk-2utg3f($l$pxcn7)ov#0-gt#n+w*uvr-6y5'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django_comments',
    'django_plotly_dash.apps.DjangoPlotlyDashConfig',
    'bootstrap4',
    'tom_alerts_dash',
    'tom_alerts',
    'tom_common',
    'tom_targets',
    'tom_observations',
    'tom_dataproducts',
    'guardian',
    'django.contrib.sites',
    'tom_setup'
]

DATA_PRODUCT_TYPES = {}

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django_plotly_dash.middleware.BaseMiddleware',
]

ROOT_URLCONF = 'tom_alerts_dash_base.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'tom_alerts_dash_base.wsgi.application'


# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


EXTRA_FIELDS = []

TOM_FACILITY_CLASSES = []


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, '_static')

STATICFILES_FINDERS = [

    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',

    'django_plotly_dash.finders.DashAssetFinder',
    'django_plotly_dash.finders.DashComponentFinder',
    'django_plotly_dash.finders.DashAppDirectoryFinder',
]


# django-plotly-dash configuration

X_FRAME_OPTIONS = 'SAMEORIGIN'

PLOTLY_COMPONENTS = [
    # Common components
    'dash_core_components',
    'dash_html_components',
    'dash_renderer',

    # django-plotly-dash components
    'dpd_components',
    # static support if serving local assets
    # 'dpd_static_support',

    # Other components, as needed
    'dash_bootstrap_components',
    'dash_table'
]



from django.urls import path

# This import is necessary for Dash to run, likely because it imports and runs the staticfiles finders
# as defined in settings.STATICFILES_FINDERS
from tom_alerts_dash import dash  # noqa
from tom_alerts_dash.views import BrokerQueryBrowseView, BrokerQueryListView

app_name = 'tom_alerts_dash'

urlpatterns = [
    path('query/list/', BrokerQueryListView.as_view(), name='list'),
    path('query/browse/', BrokerQueryBrowseView.as_view(), name='browse'),
]



from abc import abstractmethod
from importlib import import_module

from dash.dependencies import Input
from dash.exceptions import PreventUpdate
from django.conf import settings

from tom_alerts.alerts import GenericBroker


DEFAULT_ALERT_CLASSES = [
    'tom_alerts_dash.brokers.mars.MARSDashBroker',
    'tom_alerts_dash.brokers.alerce.ALeRCEDashBroker',
]


def get_service_classes():
    """
    Gets the dash broker classes available to this TOM as specified by ``TOM_ALERT_DASH_CLASSES`` in ``settings.py``.
    If none are specified, returns the default set.

    :returns: dict of broker classes, with keys being the name of the broker and values being the broker class
    :rtype: dict
    """
    try:
        TOM_ALERT_DASH_CLASSES = settings.TOM_ALERT_DASH_CLASSES
    except AttributeError:
        TOM_ALERT_DASH_CLASSES = DEFAULT_ALERT_CLASSES

    service_choices = {}
    for service in TOM_ALERT_DASH_CLASSES:
        mod_name, class_name = service.rsplit('.', 1)
        try:
            mod = import_module(mod_name)
            clazz = getattr(mod, class_name)
        except (ImportError, AttributeError):
            raise ImportError(f'Could not import {service}. Did you provide the correct path?')
        service_choices[clazz.name] = clazz
    return service_choices


def get_service_class(name):
    """
    Gets the specific dash broker class for a given broker name.

    :returns: Broker class
    :rtype: class
    """
    available_classes = get_service_classes()
    try:
        return available_classes[name]
    except KeyError:
        raise ImportError(
            '''Could not a find a broker with that name.
            Did you add it to TOM_ALERTS_DASH_CLASSES?'''
        )


class GenericDashBroker(GenericBroker):
    """
    Interface class for implementation of a Dash-compatible broker module. Please refer to the built-in ALeRCE, MARS,
    and SCIMMA Dash broker modules for implementation examples.
    """
    name = 'Generic Broker'

    def callback(self, page_current, page_size):
        """
        Broker-specific callback function triggered on broker-specific filter inputs. Accepts filter inputs and queries
        the corresponding service to return a list of flattened (single-level depth) alert dictionaries. The
        dictionaries in the return list must have keys corresponding to the ``id`` values returned by this broker's
        ``get_dash_columns()``. Method signature must correspond to the filter inputs returned by this broker's
        ``get_callback_inputs()``.

        :param page_current: The page number for the paginated alerts to display
        :type page_current: int

        :param page_size: The page size used for the pagination
        :type page_size: int

        All other args need to be specified in the concrete implementation.

        :returns: list of 1-level depth alert dicts, with keys corresponding to return value of ``get_dash_columns()``
        :rtype: list of dicts
        """
        raise PreventUpdate

    def get_callback_inputs(self):
        """
        Method that provides broker-specific inputs intended to trigger this broker's callback function. Input names
        must correspond with ``id`` properties specified in this broker's ``get_dash_filters()``.

        Default implementation provides inputs for page number and page size. This method is technically not required
        for concrete implementation, but omission would prevent an end user from filtering alerts.

        :returns: list of inputs corresponding to dash filters for this broker
        :rtype: list
        """
        return [
            Input(f'alerts-table-{self.name}', 'page_current'),
            Input(f'alerts-table-{self.name}', 'page_size')
        ]

    @abstractmethod
    def get_dash_filters(self):
        """
        Method that provides the layout and input types for the filter inputs that are available for this broker. Please
        consult the Plotly Dash documentation for specific properties of Dash components:
        https://dash.plotly.com/dash-core-components

        :returns: layout of Dash input components
        :rtype: dash_html_components.Div
        """
        pass

    @abstractmethod
    def get_dash_columns(self):
        """
        Provides the columns that will be displayed in the broker-specific Dash DataTable. Columns must follow the
        format specified in the Dash DataTable documentation: https://dash.plotly.com/datatable/reference

        :returns: columns for display in DataTable
        :rtype: list of dicts
        """
        pass

    def flatten_dash_alerts(self, alerts):
        """
        Transforms a list of alerts returned by a broker query into a list of single-level depth dictionaries for
        display in a Dash DataTable. Also handles any further transformation of the data returned by a broker query.

        Each flattened alert should also include a key/value pair of {'alert': original_alert}, to be used when
        creating a target from the alert.

        :param alerts: list of alerts from a broker query
        :type alerts: list

        :returns: list of single-level depth dicts
        :rtype: list of dicts
        """
        return alerts

    def validate_filters(self, page_current, page_size, errors_state):
        """
        Validates the input filters for a broker module. The concrete implementation of this method must accept all
        inputs returned from ``get_callback_inputs()``, as well as an ``errors_state`` as the last argument.

        :param page_current: The page number for the paginated alerts to display
        :type page_current: int

        :param page_size: The page size used for the pagination
        :type page_size: int

        All broker-specific args need to be specified in the concrete implementation.

        :param errors_state: The currently displayed errors relating to filters
        :type errors_state: list of dbc.Alert objects

        :returns: errors from validation of filters
        :rtype: list of dbc.Alert objects

        :raises: PreventUpdate
        """
        raise PreventUpdate



import logging

from astropy.time import Time
from dash.dependencies import Input
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as dhc

from tom_alerts_dash.alerts import GenericDashBroker
from tom_alerts.brokers.alerce import ALeRCEBroker, ALeRCEQueryForm, ALERCE_URL
from tom_common.templatetags.tom_common_extras import truncate_number
from tom_targets.templatetags.targets_extras import deg_to_sexigesimal

logger = logging.getLogger(__name__)


class ALeRCEDashBroker(ALeRCEBroker, GenericDashBroker):
    dash_button_clicks = 0

    def callback(self, page_current, page_size, oid, stamp_classifier, p_stamp_classifier, lc_classifier,
                 p_lc_classifier, ra, dec, radius, button_click):
        """
        ALeRCE-specific callback function for BrokerQueryBrowseView. Queries ALeRCE based on parameters from DataTable
        inputs. The callback will not trigger a query unless the "Filter" button is clicked, due to the fact that ALeRCE
        queries run very long.

        :param page_current: Currently selected page
        :type page_current: int

        :param page_size: Page size for pagination. Not currently used.
        :type page_size: int

        :param oid: ZTF objectId to search for
        :type oid: string

        :param ra: Right Ascension to use in cone search
        :type ra: string

        :param stamp_classifier: Stamp classification to filter by
        :param stamp_classifier: string

        :param p_stamp_classifier: Probability of stamp classification specified by stamp_classifier filter
        :param p_stamp_classifier: string

        :param lc_classifier: Light curve classification to filter by
        :param lc_classifier: string

        :param p_lc_classifier: Probability of light curve classification specified by lc_classifier filter
        :param p_lc_classifier: string

        :param dec: Declination to use in cone search
        :type dec: string

        :param radius: Radius to use in cone search, in degrees
        :type radius: string

        :param button_click: Number of times the filter-button has been clicked
        :param button_click: int

        :returns: list of flattened alerts
        :rtype: list of dicts

        :raises: PreventUpdate exception if some but not all cone search parameters are submitted
        """
        logger.info('Entering ALeRCE callback...')
        errors = self.validate_filters(page_current, page_size, oid, stamp_classifier, p_stamp_classifier,
                                       lc_classifier, p_lc_classifier, ra, dec, radius, button_click, [])

        # Dash does not return the state of a button, but rather the number of clicks. To determine if the callback was
        # triggered by a new button click, the broker tracks the number of clicks, and we check that it has changed
        # before querying ALeRCE.
        if not button_click or button_click == self.dash_button_clicks or errors:
            raise PreventUpdate
        else:
            self.dash_button_clicks = button_click

        form = ALeRCEQueryForm({
            'query_name': 'ALeRCE Dash Query',
            'broker': self.name,
            'oid': oid,
            'stamp_classifier': stamp_classifier,
            'p_stamp_classifier': p_stamp_classifier,
            'lc_classifier': lc_classifier,
            'p_lc_classifier': p_lc_classifier,
            'ra': ra,
            'dec': dec,
            'radius': radius
        })
        form.is_valid()
        parameters = form.cleaned_data
        parameters['page'] = page_current + 1  # Dash pagination is 0-indexed, but Skip is 1-indexed
        parameters['records_per_pages'] = page_size if page_size else 20  # 20 is the Dash default page size

        alerts = [alert for alert in self._request_alerts(form.cleaned_data)['items']]
        return self.flatten_dash_alerts(alerts)

    def get_callback_inputs(self):
        """
        Returns SCIMMA-specific inputs used to trigger callback function.

        :returns: list of inputs corresponding to dash filters
        :rtype: list
        """

        inputs = super().get_callback_inputs()
        inputs += [
            Input('oid', 'value'),
            Input('stamp_classifier', 'value'),
            Input('p_stamp_classifier', 'value'),
            Input('lc_classifier', 'value'),
            Input('p_lc_classifier', 'value'),
            Input('ra', 'value'),
            Input('dec', 'value'),
            Input('radius', 'value'),
            Input('trigger-filter-btn', 'n_clicks')
        ]
        return inputs

    def get_dash_filters(self):
        """
        Returns SCIMMA-specific filter inputs layout

        :returns: layout of Dash input components
        :rtype: dash_html_components.Div
        """
        filters = dhc.Div([
            dbc.Row([
                dbc.Col(dcc.Input(
                    id='oid',
                    type='text',
                    placeholder='Object ID',
                    debounce=True
                )),
            ], style={'padding-bottom': '10px'}),
            dbc.Row([
                dbc.Col(dcc.Dropdown(
                    id='stamp_classifier',
                    placeholder='Stamp Classifier',
                    options=[{'label': classifier[1], 'value': classifier[0]}
                             for classifier in ALeRCEQueryForm._get_stamp_classifier_choices()
                             if classifier[0] is not None]
                )),
                dbc.Col(dcc.Input(
                    id='p_stamp_classifier',
                    type='number',
                    placeholder='Stamp Classifier Probability',
                )),
                dbc.Col(dcc.Dropdown(
                    id='lc_classifier',
                    placeholder='Light Curve Classifier',
                    options=[{'label': classifier[1], 'value': classifier[0]}
                             for classifier in ALeRCEQueryForm._get_light_curve_classifier_choices()
                             if classifier[0] is not None]
                )),
                dbc.Col(dcc.Input(
                    id='p_lc_classifier',
                    type='number',
                    placeholder='Light Curve Classifier Probability',
                )),
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dcc.Input(
                    id='ra',
                    type='text',
                    placeholder='Right Ascension',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='dec',
                    type='text',
                    placeholder='Declination',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='radius',
                    type='text',
                    placeholder='Search Radius (arcseconds)',
                ), width=3)
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dbc.Button(
                    'Filter',
                    id='trigger-filter-btn',
                    outline=True,
                    color='info'
                )),
            ], style={'padding-bottom': '10px'})
        ])
        return filters

    def get_dash_columns(self):
        """
        Returns ALeRCE-specific Dash DataTable columns.

        :returns: columns for display
        :rtype: list of dicts
        """
        return [
            {'id': 'oid', 'name': 'Object ID', 'type': 'text', 'presentation': 'markdown'},
            {'id': 'meanra', 'name': 'Right Ascension', 'type': 'text'},
            {'id': 'meandec', 'name': 'Declination', 'type': 'text'},
            {'id': 'discovery_date', 'name': 'Discovery Date', 'type': 'datetime'},
            {'id': 'classifier', 'name': 'Class', 'type': 'text'},
            {'id': 'class', 'name': 'Classifier Type', 'type': 'text'},
            {'id': 'probability', 'name': 'Classifier Probability', 'type': 'text'},
        ]

    def flatten_dash_alerts(self, alerts):
        """
        Transforms alerts returned by ALeRCE into a Dash DataTable format. Adds an embedded link to the original alert.
        Converts decimal degrees to sexagesimal. Truncates decimals to 4 places. Converts MJD value to datetime.
        Includes light curve classifier if it exists, and stamp classifier otherwise. Displays classifier name instead
        of classifier number.

        :param alerts: dict of alerts from ALeRCE
        :type alerts: dict of dicts

        :returns: flattened alerts
        :rtype: list of dicts
        """
        flattened_alerts = []
        for alert in alerts:
            url = f'{ALERCE_URL}/object/{alert["oid"]}'
            flattened_alerts.append({
                'oid': f'[{alert["oid"]}]({url})',
                'meanra': deg_to_sexigesimal(alert['meanra'], 'hms') if alert['meanra'] else None,
                'meandec': deg_to_sexigesimal(alert['meandec'], 'dms') if alert['meandec'] else None,
                'discovery_date': Time(alert['firstmjd'], format='mjd', scale='utc').to_datetime(),
                'class': alert['class'],
                'classifier': alert['classifier'],
                'probability': truncate_number(alert['probability']),
                'alert': alert
            })
        return flattened_alerts

    def validate_filters(self, page_current, page_size, oid, stamp_classifier, p_stamp_classifier, lc_classifier,
                         p_lc_classifier, ra, dec, radius, button_click, errors_state):
        """
        Validates the input filters for ALeRCE. Returns an error if one, but not all, of RA, Dec, and radius are
        submitted for cone search. Returns any errors generated by form validation.

        :param page_current: The page number for the paginated alerts to display
        :type page_current: int

        :param page_size: The page size used for the pagination
        :type page_size: int

        :param oid: ZTF objectId to search for
        :type oid: string

        :param stamp_classifier: Stamp classification to filter by
        :param stamp_classifier: string

        :param p_stamp_classifier: Probability of stamp classification specified by stamp_classifier filter
        :param p_stamp_classifier: string

        :param lc_classifier: Light curve classification to filter by
        :param lc_classifier: string

        :param p_lc_classifier: Probability of light curve classification specified by lc_classifier filter
        :param p_lc_classifier: string

        :param ra: Right Ascension to use in cone search
        :type ra: string

        :param dec: Declination to use in cone search
        :type dec: string

        :param radius: Radius to use in cone search, in degrees
        :type radius: string

        :param button_click: Number of times the filter-button has been clicked
        :param button_click: int

        :param errors_state: The currently displayed errors relating to filters
        :type errors_state: list of dbc.Alert objects

        :returns: errors from validation of filters
        :rtype: list of dbc.Alert objects
        """
        errors = []

        if not button_click or button_click == self.dash_button_clicks:
            raise PreventUpdate

        form = ALeRCEQueryForm({
            'query_name': 'ALeRCE Dash Query',
            'broker': self.name,
            'oid': oid,
            'stamp_classifier': stamp_classifier,
            'p_stamp_classifier': p_stamp_classifier,
            'lc_classifier': lc_classifier,
            'p_lc_classifier': p_lc_classifier,
            'ra': ra,
            'dec': dec,
            'radius': radius
        })
        form.is_valid()

        for field, field_errors in form.errors.items():
            for field_error in field_errors.get_json_data():
                errors.append(f'{field}: {field_error["message"]}')

        for error in errors:
            errors_state.append(dbc.Alert(error, dismissable=True, is_open=True, duration=5000, color='warning'))

        return errors_state



from datetime import datetime
import logging

from dash.dependencies import Input
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import dash_html_components as dhc
import dash_core_components as dcc

from tom_alerts_dash.alerts import GenericDashBroker
from tom_alerts.brokers.mars import MARSBroker, MARSQueryForm, MARS_URL
from tom_common.templatetags.tom_common_extras import truncate_number
from tom_targets.templatetags.targets_extras import deg_to_sexigesimal

logger = logging.getLogger(__name__)


class MARSDashBroker(MARSBroker, GenericDashBroker):
    dash_button_clicks = 0

    def callback(self, page_current, page_size, objectId, cone_ra, cone_dec, cone_radius, magpsf_lte, rb_gte,
                 start_date, end_date, button_click):
        """
        MARS-specific callback function for BrokerQueryBrowseView. Queries MARS based on parameters from DataTable
        inputs. The callback will not modify any bound components if one or more, but not all, cone search inputs are
        submitted, or if the form validation fails.

        :param page_current: Currently selected page
        :type page_current: int

        :param page_size: Page size for pagination. Not currently used.
        :type page_size: int

        :param objectId: ZTF objectId to search for
        :type objectId: string

        :param cone_ra: Right Ascension to use in cone search
        :type cone_ra: string

        :param cone_dec: Declination to use in cone search
        :type cone_dec: string

        :param cone_radius: Radius to use in cone search, in degrees
        :type cone_radius: string

        :param magpsf_lte: Maximum magnitude to filter by
        :param magpsf_lte: string

        :param rb_gte: Minimum real-bogus score to filter by
        :param rb_gte: string

        :param start_date: Earliest date to filter by
        :param start_date: string

        :param end_date: Latest date to filter by
        :param end_date: string

        :param button_click: Number of times the filter-button has been clicked
        :param button_click: int

        :returns: list of flattened alerts
        :rtype: list of dicts

        :raises: PreventUpdate exception if some but not all cone search parameters are submitted
        """
        logger.info('Entering MARS callback...')
        errors = self.validate_filters(page_current, page_size, objectId, cone_ra, cone_dec, cone_radius, magpsf_lte,
                                       rb_gte, start_date, end_date, button_click, [])

        if not button_click or button_click == self.dash_button_clicks or errors:
            raise PreventUpdate
        else:
            self.dash_button_clicks = button_click

        cone_search = ''
        if all([cone_ra, cone_dec, cone_radius]):
            cone_search = ','.join([cone_ra, cone_dec, cone_radius])
        form = MARSQueryForm({
            'query_name': 'dash query',
            'broker': self.name,
            'objectId': objectId,
            'magpsf__lte': magpsf_lte,
            'rb__gte': rb_gte,
            'time__gt': start_date,
            'time__lt': end_date,
            'cone': cone_search
        })
        form.is_valid()

        parameters = form.cleaned_data
        parameters['page'] = page_current + 1  # Dash pagination is 0-indexed, but MARS is 1-indexed

        alerts = self._request_alerts(parameters)['results']
        return self.flatten_dash_alerts(alerts)

    def get_callback_inputs(self):
        """
        Returns MARS-specific inputs used to trigger callback function.

        :returns: list of inputs corresponding to dash filters
        :rtype: list
        """
        inputs = super().get_callback_inputs()
        inputs += [
            Input('mars-objname-search', 'value'),
            Input('mars-cone-ra', 'value'),
            Input('mars-cone-dec', 'value'),
            Input('mars-cone-radius', 'value'),
            Input('mars-magpsf-max', 'value'),
            Input('mars-rb-min', 'value'),
            Input('mars-date-filter', 'start_date'),
            Input('mars-date-filter', 'end_date'),
            Input('mars-trigger-filter-btn', 'n_clicks')
        ]
        return inputs

    def get_dash_filters(self):
        """
        Returns MARS-specific filter inputs layout

        :returns: layout of Dash input components
        :rtype: dash_html_components.Div
        """
        filters = dhc.Div([
            dbc.Row([
                dbc.Col(dcc.Input(
                    id='mars-objname-search',
                    type='text',
                    placeholder='Object Name Search',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='mars-magpsf-max',
                    type='number',
                    placeholder='Magnitude Maximum',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='mars-rb-min',
                    type='number',
                    placeholder='Real-Bogus Minimum',
                    debounce=True
                ), width=3)
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dcc.Input(
                    id='mars-cone-ra',
                    type='text',
                    placeholder='Right Ascension',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='mars-cone-dec',
                    type='text',
                    placeholder='Declination',
                    debounce=True
                ), width=3),
                dbc.Col(dcc.Input(
                    id='mars-cone-radius',
                    type='text',
                    placeholder='Radius',
                    debounce=True
                ), width=3)
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dcc.DatePickerRange(
                    id='mars-date-filter',
                    min_date_allowed=datetime(2018, 1, 1),
                    initial_visible_month=datetime.now(),
                    clearable=True
                ))
            ], style={'padding-bottom': '10px'}, justify='start'),
            dbc.Row([
                dbc.Col(dbc.Button(
                    'Filter',
                    id='mars-trigger-filter-btn',
                    outline=True,
                    color='info'
                )),
            ], style={'padding-bottom': '10px'})
        ])
        return filters

    def get_dash_columns(self):
        """
        Returns MARS-specific Dash DataTable columns.

        :returns: columns for display
        :rtype: list of dicts
        """
        return [
            {'id': 'objectId', 'name': 'Name', 'type': 'text', 'presentation': 'markdown'},
            {'id': 'ra', 'name': 'Right Ascension', 'type': 'text'},
            {'id': 'dec', 'name': 'Declination', 'type': 'text'},
            {'id': 'magpsf', 'name': 'Magnitude', 'type': 'text'},
            {'id': 'rb', 'name': 'Real-Bogus Score', 'type': 'text'},
        ]

    def flatten_dash_alerts(self, alerts):
        """
        Transforms alerts returned by MARS into a Dash DataTable format. Adds an embedded link to the original alert.
        Converts decimal degrees to sexagesimal. Truncates decimals to 4 places.

        :param alerts: list of alerts from MARS
        :type alerts: list of dicts

        :returns: flattened alerts
        :rtype: list of dicts
        """
        flattened_alerts = []
        for alert in alerts:
            url = f'{MARS_URL}/{alert["lco_id"]}/'
            flattened_alerts.append({
                'objectId': f'[{alert["objectId"]}]({url})',
                'ra': deg_to_sexigesimal(alert['candidate']['ra'], 'hms'),
                'dec': deg_to_sexigesimal(alert['candidate']['dec'], 'dms'),
                'magpsf': truncate_number(alert['candidate']['magpsf']),
                'rb': truncate_number(alert['candidate']['rb']),
                'alert': alert
            })
        return flattened_alerts

    def validate_filters(self, page_current, page_size, objectId, cone_ra, cone_dec, cone_radius, magpsf_lte, rb_gte,
                         start_date, end_date, button_click, errors_state):
        """
        Validates the input filters for MARS. Returns an error if one, but not all, of RA, Dec, and radius are submitted
        for cone search. Returns any errors generated by form validation.

        :param page_current: The page number for the paginated alerts to display
        :type page_current: int

        :param page_size: The page size used for the pagination
        :type page_size: int

        :param objectId: ZTF objectId to search for
        :type objectId: string

        :param cone_ra: Right Ascension to use in cone search
        :type cone_ra: string

        :param cone_dec: Declination to use in cone search
        :type cone_dec: string

        :param cone_radius: Radius to use in cone search, in degrees
        :type cone_radius: string

        :param magpsf_lte: Maximum magnitude to filter by
        :param magpsf_lte: string

        :param rb_gte: Minimum real-bogus score to filter by
        :param rb_gte: string

        :param errors_state: The currently displayed errors relating to filters
        :type errors_state: list of dbc.Alert objects

        :returns: errors from validation of filters
        :rtype: list of dbc.Alert objects
        """
        errors = []

        cone_search = ''
        if any([cone_ra, cone_dec, cone_radius]):
            if all([cone_ra, cone_dec, cone_radius]):
                cone_search = ','.join([cone_ra, cone_dec, cone_radius])
            else:
                errors.append('All of RA, Dec, and Radius are required for a cone search.')

        form = MARSQueryForm({
            'query_name': 'dash query',
            'broker': self.name,
            'objectId': objectId,
            'magpsf__lte': magpsf_lte,
            'rb__gte': rb_gte,
            'time__gt': start_date,
            'time__lt': end_date,
            'cone': cone_search
        })
        form.is_valid()

        for field, field_errors in form.errors.items():
            for field_error in field_errors.get_json_data():
                errors.append(f'{field}: {field_error["message"]}')

        for error in errors:
            errors_state.append(dbc.Alert(error, dismissable=True, is_open=True, duration=5000, color='warning'))

        return errors_state



import factory
from faker import Faker

from tom_targets.models import Target

fake = Faker()


class SiderealTargetFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Target

    name = factory.Faker('pystr')
    type = Target.SIDEREAL
    ra = factory.Faker('pyfloat', min_value=-90, max_value=90)
    dec = factory.Faker('pyfloat', min_value=-90, max_value=90)
    epoch = factory.Faker('pyfloat')
    pm_ra = factory.Faker('pyfloat')
    pm_dec = factory.Faker('pyfloat')


def create_mars_alert(ra=None, dec=None, magpsf=None, rb=None):
    return {
        'avro': fake.uri(),
        'candid': fake.pyint(min_value=1000000000000000000, max_value=9999999999999999999),
        'objectId': fake.pystr_format(string_format='ZTF##???????'),
        'lco_id': fake.pyint(min_value=100000000, max_value=999999999),
        'candidate': {
            'ra': ra if ra else fake.pyfloat(min_value=0, max_value=360),
            'dec': dec if dec else fake.pyfloat(min_value=0, max_value=360),
            'magpsf': magpsf if magpsf else fake.pyfloat(min_value=12, max_value=22),
            'rb': rb if rb else fake.pyfloat(min_value=0, max_value=1),
            'drb': fake.pyfloat(min_value=0, max_value=1)
        }
    }


def create_scimma_alert(ra=None, dec=None, rank=None):
    return {
        'alert_identifier': fake.pystr_format(string_format='S######y_X##'),
        'right_ascension_sexagesimal': ra if ra else fake.pystr_format(string_format='##:##:##.###'),
        'declination_sexagesimal': dec if dec else fake.pystr_format(string_format='##:##:##.###'),
        'topic': fake.pystr(max_chars=5),
        'message': {
            'rank': rank if rank else fake.pyint(min_value=1, max_value=4),
            'event_trig_num': fake.pystr_format(string_format='S######')
        },
        'extracted_fields': {
            'counterpart_identifier': fake.pystr_format(string_format='#??? ?######.#+######'),
            'comment_warnings': fake.pystr(),
        }
    }


def create_alerce_alert(meanra=None, meandec=None, firstmjd=None, lastmjd=None, class_name=None, classifier=None,
                        probability=None):
    return {'oid': fake.pystr_format(string_format='ZTF##???????', letters='abcdefghijklmnopqrstuvwxyz'),
            'meanra': meanra if meanra else fake.pyfloat(min_value=0, max_value=360),
            'meandec': meandec if meandec else fake.pyfloat(min_value=0, max_value=360),
            'firstmjd': firstmjd if firstmjd else fake.pyfloat(min_value=56000, max_value=59000, right_digits=1),
            'lastmjd': lastmjd if lastmjd else fake.pyfloat(min_value=56000, max_value=59000, right_digits=1),
            'class': class_name if class_name else fake.pystr_format(string_format='????????',
                                                                     letters='ABCDEFGHIJKLMNOPQRSTUVWXYZ'),
            'classifier': classifier if classifier else fake.pystr_format(string_format='???',
                                                                          letters='ABCDEFGHIJKLMNOPQRSTUVWXYZ'),
            'probability': probability if probability else fake.pyfloat(min_value=0, max_value=1)}



import logging

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as dhc
from dash_table import DataTable
from django_plotly_dash import DjangoDash
from django.shortcuts import reverse

from tom_alerts_dash.alerts import get_service_class, get_service_classes

# This module creates the browseable alert tables for the supported brokers. It does so by creating a Dash container for
# each registered broker in settings.py. The containers include two messages containers, a create-targets button, a set
# of broker-specific filter inputs, and a DataTable. The containers are set to {display: none;} on load with no alerts
# in them. A callback is registered that listens to a dropdown allowing the user to select a broker--when the broker
# selection changes, the corresponding broker container is displayed. Additionally, a callback is registered for each
# broker, with the broker-specific filters as the inputs, and the broker-specific DataTable as the output. Finally, two
# callbacks are registered for each broker for error or success messages. The first uses the broker-specific filters as
# the inputs, and one of the two broker-specific messages containers as the output. The second uses the broker-specific
# create-targets button as the input, and the other broker-specific messages container as the output.

logger = logging.getLogger(__name__)

app = DjangoDash('BrokerQueryListViewDash', external_stylesheets=[dbc.themes.BOOTSTRAP], add_bootstrap_links=True)


def create_targets(create_targets, selected_rows, row_data, broker_state, messages_state):
    """
    Create TOM Toolkit target objects for each selected target for the current broker. Callback is triggered by a click
    of the broker-specific create-targets-{broker_name} button. Upon clicking, the callback gets the current-selected
    rows in the broker-specific DataTable and calls ``tom_alerts.alerts.to_target`` on each one.

    This fires on page load, but should not. However, the ``prevent_initial_call`` kwargs does not appear to work in
    django-plotly-dash.

    :param create_targets: Number of times the create-targets button has been clicked.
    :type create_targets: int

    :param selected_rows: indices of rows selected in the DataTable. As a State value, this does not trigger callback.
    :type selected_rows: list

    :param row_data: Data currently displayed in the DataTable. As a State value, this does not trigger callback.
    :type row_data: list of dicts

    :param broker_state: Currently selected broker. As a State value, this does not trigger callback.
    :type broker_state: str

    :param messages_state: Currently displayed messages. As a State value, this does not trigger callback.
    :type messages_state: list of dbc.Alert object
    """
    logger.info(f'Entering create targets callback for broker: {broker_state}')
    # Ensure the create-targets button has actually been clicked and that there are selected rows
    if create_targets and selected_rows:
        broker_class = get_service_class(broker_state)()
        messages = messages_state
        for row in selected_rows:
            try:
                target = broker_class.to_target(row_data[row]['alert'])  # Get the data for each selected row
                target_url = reverse('targets:detail', kwargs={'pk': target.id})
                messages.append(
                    dbc.Alert(['Successfully created ', dhc.A('View Target', href=target_url)],
                              color='success', dismissable=True, duration=5000, is_open=True)
                )
            except Exception as e:
                logger.error(f'Unable to create target from alert {row_data[row]["alert"]} due to exception {e}.')
                messages.append(
                    dbc.Alert('Unable to create target from alert.',  # TODO: how to give the alert name?
                              color='danger', is_open=True, dismissable=True, duration=5000)
                )

        return messages
    else:
        raise PreventUpdate


def create_broker_callbacks():
    """
    Add all broker-specific callbacks to the app callbacks on init, and construct the alerts table
    dynamically, with a different id depending on the broker. As there's no way to remove callbacks,
    this is the only way to support different callbacks per broker.

    There are three broker-specific callbacks per broker. The first is a callback that fires on a change in any
    broker-specific inputs and updates the data in the broker-specific DataTable. The second fires on a click of the
    broker-specific create-targets button and updates the broker-specific messages container in order to convey success
    or failure of target creation. The third fires on any change in broker-specific inputs and validates the inputs,
    then returns Alert objects to display to the user any validation errors.
    """

    for class_name in get_service_classes().keys():
        broker_class = get_service_class(class_name)()
        table_callback = app.callback(  # Create the broker-specific filters callback
            Output(f'alerts-table-{class_name}', 'data'),
            broker_class.get_callback_inputs()
        )
        table_callback(broker_class.callback)  # Instantiate the broker-specific filters callback

        filter_validation_callback = app.callback(  # Create the broker-specific filter validation callback
            Output(f'messages-filters-{class_name}', 'children'),
            broker_class.get_callback_inputs(),
            [State(f'messages-filters-{class_name}', 'children')]
        )
        filter_validation_callback(broker_class.validate_filters)

        create_targets_callback = app.callback(  # Create the broker-specific create-targets callback
            Output(f'messages-targets-{class_name}', 'children'),
            [Input(f'create-targets-btn-{class_name}', 'n_clicks')],
            [State(f'alerts-table-{class_name}', 'derived_virtual_selected_rows'),
             State(f'alerts-table-{class_name}', 'derived_virtual_data'),
             State('broker-state', 'value'),
             State(f'messages-targets-{class_name}', 'children')]
        )
        create_targets_callback(create_targets)  # Create the broker-specific create-targets callback


def create_broker_container(broker):
    """
    This method creates the container with the broker-specific components. It is hidden by default. The components are
    a redirection container, a series of filter input components, a create-targets button, and a Dash DataTable. Each
    component id includes the name of the broker, in order to distinguish it for use in a specific callback function.

    :param broker: The name of the broker class for which to create a container
    :type broker: str

    :returns: The container with the redirection, filter inputs, button, and DataTable
    :rtype: dhc.Div
    """
    broker_class = get_service_class(broker)()
    return dhc.Div(children=[
        dcc.Loading(children=[
            dhc.Div(
                broker_class.get_dash_filters()
            ),
            dhc.Div(
                dhc.P(
                    dbc.Button(
                        'Create targets from selected',
                        id=f'create-targets-btn-{broker}',
                        outline=True,
                        color='info'
                    ),
                )
            ),
            DataTable(
                id=f'alerts-table-{broker}',
                columns=broker_class.get_dash_columns(),
                data=[],
                row_selectable='multi',
                page_current=0,
                page_size=20,
                page_action='custom',
                css=[
                    {'selector': '.dash-cell-value', 'rule': 'backgroundColor: blue;'}
                ],
                style_cell_conditional=[
                    {
                        # 'if': {'column_id': 'objectId'}, 'backgroundColor': 'blue'
                    }
                ],
                style_cell={
                    'textAlign': 'right'
                },
                style_data_conditional=[
                    {
                        'if': {'row_index': 'odd'},
                        'backgroundColor': 'rgb(233, 243, 256)'
                    },
                ],
                style_data={'font-family': 'Helvetica Neue, Helvetica, Arial, sans-serif'},
                style_filter={
                    'font-family': 'Helvetica Neue, Helvetica, Arial, sans-serif',
                    'backgroundColor': 'rgb(256, 233, 233)'
                },
                style_header={
                    'backgroundColor': 'rgb(213, 223, 242)',
                    'font-family': 'Helvetica Neue, Helvetica, Arial, sans-serif',
                    'fontWeight': 'bold'
                }
            )
        ], id=f'alerts-loading-container-{broker}'),
    ], id=f'alerts-container-{broker}', style={'display': 'none'})


@app.callback(
    [Output('broker-state', 'value'), Output('page-header', 'children')] +
    [Output(f'alerts-container-{clazz}', 'style') for clazz in get_service_classes().keys()],
    [Input('broker-selection', 'value')],
    [State('broker-state', 'value')]
)
def broker_selection_callback(broker_selection, broker_state):
    """
    Callback triggered by a selection of the broker dropdown. The callback also takes the previously selected broker.
    The outputs are the broker state container, the header displaying which broker alerts are being viewed,
    and an output bound to the style property of each broker container.

    If the broker selection did not change from the previously selected broker, no update occurs.

    If the broker selection did change, the following events occur:
    - The newly selected broker is added to the return values in order to update the broker-state component
    - A new dhc.H3 element with the correct broker display name is added to the return values
    - A style property is added to the return values for each broker container. The style property is
      {'display': 'none'} for all brokers save the selected broker, which instead is {'display': 'block'}. This will
      hide all containers except the one for the selected broker.

    :param broker_selection: The newly selected broker
    :type broker_selection: str

    :param broker_state: The previously selected broker. As a State value, a value change does not trigger the callback.
    :type broker_state: str

    :returns: The value of the newly selected broker
    :rtype: str

    :returns: A header showing the name of the newly selected broker
    :rtype: dash_html_component.H3

    :returns: A CSS style dictionary for each broker, either {'display': 'none'} or {'display': 'block'}
    :rtype: dict

    :raises: PreventUpdate when the newly selected broker does not change
    """
    callback_return_values = ()
    if broker_selection and broker_selection != broker_state:  # Broker selection has changed

        # Modify page header to display correct broker name
        page_header = dhc.H3(f'{broker_selection} Alerts')

        # Add the newly selected broker and new page_header to the return tuple
        callback_return_values += (broker_selection, page_header)

        # Hide all DataTables other than the one that corresponds with the selected broker
        for clazz in get_service_classes().keys():
            if broker_selection == clazz:  # newly selected broker should be displayed
                callback_return_values += ({'display': 'block'},)
            else:  # all other brokers should be hidden
                callback_return_values += ({'display': 'none'},)

        return callback_return_values
    else:  # Broker selection has not changed from previous value
        raise PreventUpdate  # Don't update any components


create_broker_callbacks()

app.layout = dbc.Container([
    dhc.Div(
        # Messages containers for validation messages related to filter inputs
        [dhc.Div(children=[], id=f'messages-filters-{class_name}') for class_name in get_service_classes().keys()] +
        # Messages containers for validation messages related to target creation
        [dhc.Div(children=[], id=f'messages-targets-{class_name}') for class_name in get_service_classes().keys()] +
        [
            dhc.Div(  # Create an initial header. This div will be replaced by the broker_selection callback
                dhc.H3('View Alerts for a Broker'),
                id='page-header'
            ),
            dhc.Div(children=[
                # Hidden component to store the currently selected broker. This is used for the create_targets callback.
                dcc.Input(id='broker-state', type='hidden', value=''),
                dhc.P(
                    dcc.Dropdown(  # Dropdown component to select the active broker
                        id='broker-selection',
                        placeholder='Select Broker',
                        options=[{'label': clazz, 'value': clazz} for clazz in get_service_classes().keys()]
                    )
                )
            ]),
            dhc.Div(  # Creates a container for each broker
                children=[create_broker_container(class_name) for class_name in get_service_classes().keys()],
            ),
        ]
    )
])



from django.views.generic import TemplateView

from tom_alerts.views import BrokerQueryListView


class BrokerQueryBrowseView(TemplateView):
    template_name = 'tom_alerts_dash/brokerquery_browse.html'


class BrokerQueryListView(BrokerQueryListView):
    template_name = 'tom_alerts_dash/brokerquery_list.html'
